unit U_llvl;

{ author:    Michael Hieke, 1997, 1998
  e-mail:    MHieke42@aol.com             }


{$IFDEF WIN32}
  szWarning = 'Don''t try this at home!';
{$ELSE}

{$I settings.inc}

interface

uses WinTypes, WinProcs, Messages, Toolhelp, PrcHook, Classes;

const
  WM_MODULELISTCHANGED     =  WM_USER;
  WM_REMOVEDLLS            =  WM_USER + 1;


{******************************************************************************}
{ NotifyRegister() etc.                                                        }
{******************************************************************************}

type
  PNFYEvent = ^TNFYEvent;
  TNFYEvent = record
    wID: word;                   { notification code }
    hMod: THandle;               { handle of added or removed module }
    atomMsg: TAtom;              { handle of atom containing message }
  end;

procedure NotificationStart(wWndPost: HWND);
procedure NotificationStop;

function GetNextPendingEvent: PNFYEvent;

const
    { Use the event model to show private messages. This way the order of
      API calls and kernel notifications will be preserved in the output.
      Define some values above the NFY_xxx constants. }
    NFY_MH_DEBUGOUT       =   1000;
    NFY_MH_LOADLIBRARY    =   1001;
    NFY_MH_FREELIBRARY    =   1002;

{******************************************************************************}
{ intercepted API calls, new window procedure for TAlignPalette                }
{******************************************************************************}

procedure APIInterceptionStart;
procedure APIInterceptionStop;


{******************************************************************************}

{$IFDEF DEBUG}
function GetDelphiTask: THandle;
{$ENDIF}

implementation

uses SysUtils;

{$IFDEF DEBUG}
var
  szODS: array[0..200] of char;

type
  TDescString = string[40];

function EventDescr(wID: word): TDescString;
begin
  case wID of
    NFY_STARTTASK: Result:= 'TASK START';
    NFY_EXITTASK: Result:= 'TASK EXIT';
    NFY_STARTDLL: Result:= 'DLL LOAD';
    NFY_DELMODULE: Result:= 'MODULE UNLOAD';
    else Result:= '';
  end;
end;
{$ENDIF}

{******************************************************************************}
{ NotifyRegister() etc.                                                        }
{******************************************************************************}

const
    { list for pending notification events }
  gaPendingEvents: TList = nil;
  pfnNotifyCallback: TFarProc = nil;
    { post message to application main window to process pending events }
  ghPostWnd: HWND = 0;
  gbMsgPosted: boolean = FALSE;
    { which task do we want to spy upon? }
  ghTaskDelphi: THandle = 0;

{$IFDEF DEBUG}
function GetDelphiTask: THandle;
begin
  Result:= ghTaskDelphi;
end;
{$ENDIF}

{***}

procedure AddEvent(pEvent: PNFYEvent);
begin
  gaPendingEvents.Add(pEvent);
    { post message to main window }
  if (not gbMsgPosted and IsWindow(ghPostWnd)) then begin
    gbMsgPosted:= TRUE;
    PostMessage(ghPostWnd, WM_MODULELISTCHANGED, 0, 0);
  end;
end;

function NotifyCallback(wID: word; dwData: longint): BOOL; export;
var wTaskID, hMod: THandle;
    pEvt: PNFYEvent;
begin
    { gets called in another task's context }
  asm
    MOV    AX, SEG @DATA
    MOV    DS, AX
  end;
    { keep them other progs goin' }
  Result:= FALSE;
  if (ghPostWnd = 0) then begin
    OutputDebugString('0'#10);
    exit;
  end;

  wTaskID:= GetCurrentTask;
  if (((ghTaskDelphi = 0) and (wID = NFY_STARTTASK)) or (ghTaskDelphi = wTaskID))
  then begin
    hMod:= 0;
    case wID of
        { task was started }
      NFY_STARTTASK:
        if (ghTaskDelphi = 0) then begin
          ghTaskDelphi:= wTaskID;
            { use the task database entry to get the module handle
              after some trying in PASCAL and strange GPF's I took the easy way out: }
          asm
            MOV    AX, wTaskID
            MOV    ES, AX
              { see PIETREK: "What's the difference between Windows and Tasks
                              in Windows 3.1?", MSDN or MSJ October 1993 }
            MOV    AX, ES:[01Eh]
            MOV    hMod, AX
          end;
        end;
{$IFDEF DEBUG}
        { task has ended }
      NFY_EXITTASK:
        asm
          MOV    AX, wTaskID
          MOV    ES, AX
          MOV    AX, ES:[01Eh]
          MOV    hMod, AX
        end;
{$ENDIF}
        { new dll loaded }
      NFY_STARTDLL:
        hMod:= PNFYStartDLL(dwData)^.hModule;
        { dll or main module of task unloaded }
      NFY_DELMODULE:
        hMod:= LoWord(dwData);
    end;
    if ((hMod <> 0) and (gaPendingEvents.Count < 1000)) then begin
      New(pEvt);
      if Assigned(pEvt) then begin
        pEvt^.wID:= wID;
        pEvt^.hMod:= hMod;
        AddEvent(pEvt);
      end;
    end;
  end;
end;

procedure NotificationStart(wWndPost: HWND);
begin
{$IFDEF DEBUG}
  OutputDebugString('start notification'#10);
{$ENDIF}
    { Don't use any API funcs in callback except those in TOOLHELP.DLL.
      Store Events in list and post msg to main window }
  if (gaPendingEvents = nil) then gaPendingEvents:= TList.Create;
  if IsWindow(wWndPost) then ghPostWnd:= wWndPost
                        else ghPostWnd:= 0;
    { create thunk for callback and register it }
  pfnNotifyCallback:= MakeProcInstance(@NotifyCallback, HInstance);
  NotifyRegister(0, TNotifyCallBack(pfnNotifyCallback), NF_NORMAL);
end;

procedure NotificationStop;
var i: integer;
begin
    { unregister callback and free thunk }
  NotifyUnregister(0);
  FreeProcInstance(pfnNotifyCallback);
    { free pending event objects }
  if (gaPendingEvents.Count > 0) then begin
{$IFDEF DEBUG}
    StrFmt(szODS, '%d event(s) left unprocessed'#10, [gaPendingEvents.Count]);
    OutputDebugString(szODS);
{$ENDIF}
    for i:= 0 to Pred(gaPendingEvents.Count) do Dispose(PNFYEvent(gaPendingEvents[i]));
  end;
  gaPendingEvents.Free;
{$IFDEF DEBUG}
  OutputDebugString('notification stopped'#10);
{$ENDIF}
end;

function GetNextPendingEvent: PNFYEvent;
begin
  if (gaPendingEvents.Count > 0) then begin
    Result:= gaPendingEvents[0];
    gaPendingEvents.Delete(0);
  end else begin
    gbMsgPosted:= FALSE;
    Result:= nil;
  end;
end;


{******************************************************************************}
{ intercepted API calls                                                        }
{******************************************************************************}

const
    { constants from WIN32 windows.h missing in 16-bit }
  WS_EX_TOOLWINDOW      = $80;
  SM_CYSMCAPTION        = 51;
    { don't touch this form }
  AppBuilderClass       = 'TAppBuilder';
    { new height for this form }
  AlignPaletteClass     = 'TAlignPalette';


const
    { The CreateWindowEx() call is used throughout the VCL to create visible
      components. All nonmodal forms except the main form TAppBuilder get
      the WS_EX_TOOLWINDOW extended style to remove them from the Windows95
      task bar. Is there any better window style to do this without removing
      the system menu and the minimize / maximize buttons??? }
  hookCreateWndEx: NPHookChild = 0;
  lpfnCreateWndEx: TFarProc = nil;


    { TAlignPalette needs a modified window procedure.
      WM_GETMINMAXINFO returns hardcoded values which screw up all changes
      to SetWindowPos() parameters. A new window procedure is the best
      in terms of performance, because no other windows are touched as with
      hooking an API call.
      To remove the extra height the height values are decremented for
      (GetSystemMetrics(SM_CYCAPTION) - GetSystemMetrics(SM_CYSMCAPTION) - 1) }
  lpfnOldWndProc: TFarProc = nil;
  lpfnNewWndProc: TFarProc = nil;
  ghAlignWindow: HWND = 0;


    { Keep track of LoadLibrary() and FreeLibrary() calls, because only
      explicitely loaded DLL's should be unloaded later on }
  hookLoadLibrary: NPHookChild = 0;
  hookFreeLibrary: NPHookChild = 0;
  lpfnLoadLibrary: TFarProc = nil;
  lpfnFreeLibrary: TFarProc = nil;

{***}

function _SafePChar(pszText: PChar): PChar;
const pszNull: PChar = #0;
begin
  if Assigned(pszText) then Result:= pszText
                       else Result:= pszNull;
end;

function CreateWndExProc(ExStyle: LongInt; ClassName, WindowName: PChar;
                         Style: LongInt; X, Y, Width, Height: Integer;
                         WndParent: HWnd; Menu: HMenu; Instance: THandle;
                         Param: Pointer): HWnd; export;
{$IFDEF DEBUG}
var pEvt: PNFYEvent;
    bOutput: boolean;
{$ENDIF}
begin
  asm
    MOV    AX, SEG @DATA
    MOV    DS, AX
  end;

  if (GetCurrentTask = ghTaskDelphi) then begin
      { only for top level, nonmodal forms }
    if (Style and (WS_CHILD or DS_MODALFRAME) = 0) then
      { skip DELPHI's main form }
    if (StrComp(_SafePChar(ClassName), AppBuilderClass) <> 0) then begin
{$IFDEF DEBUG}
      bOutput:= TRUE;
{$ENDIF}
      ExStyle:= ExStyle or WS_EX_TOOLWINDOW;
    end;
{$IFDEF DEBUG}
  end else
    bOutput:= FALSE;
{$ELSE}
  end;
{$ENDIF}
    { call the real CreateWindowEx() }
  ProcUnhook(hookCreateWndEx);
  Result:= CreateWindowEx(ExStyle, ClassName, WindowName, Style,
                          X, Y, Width, Height, WndParent, Menu, Instance, Param);
  ProcHook(hookCreateWndEx);

{$IFDEF DEBUG}
  if bOutput then begin
    StrFmt(szODS, 'CreateWindowEx(): %s "%s"  (%d, %d, %d, %d): %4xh'#10,
                  [ClassName, _SafePChar(WindowName), X, Y, Width, Height, Result]);
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_DEBUGOUT;
      pEvt^.atomMsg:= GlobalAddAtom(szODS);
      AddEvent(pEvt);
    end;
  end;
{$ENDIF}

    { if valid TAlignPalette form handle subclass it }
  if ((GetCurrentTask = ghTaskDelphi) and (ghAlignWindow = 0) and
      (Result <> 0) and (StrIComp(_SafePChar(ClassName), AlignPaletteClass) = 0))
  then begin
    ghAlignWindow:= Result;
    lpfnOldWndProc:= TFarProc(GetWindowLong(ghAlignWindow, GWL_WNDPROC));
    SetWindowLong(ghAlignWindow, GWL_WNDPROC, longint(lpfnNewWndProc));
{$IFDEF DEBUG}
    StrFmt(szODS, 'window %4xh: changed window proc from %p to %p'#10,
                  [ghAlignWindow, lpfnOldWndProc, lpfnNewWndProc]);
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_DEBUGOUT;
      pEvt^.atomMsg:= GlobalAddAtom(szODS);
      AddEvent(pEvt);
    end;
{$ENDIF}
  end;
end;

{***}

function AlignWindowProc(Wnd: HWND; Msg, wParam: word; lParam: longint): longint; export;
var wDelta: integer;
{$IFDEF DEBUG}
    pEvt: PNFYEvent;
{$ENDIF}
begin
    { force the right DS, call saved window procedure as default }
  asm
    MOV     AX, SEG @DATA
    MOV     DS, AX

    PUSH    Wnd
    PUSH    Msg
    PUSH    wParam
    PUSH    lParam.WORD[2]
    PUSH    lParam.WORD[0]
    CALL    lpfnOldWndProc
    MOV     Result.WORD[0], AX
    MOV     Result.WORD[2], DX
  end;
    { don't restore old window procedure on cleanup if window got destroyed }
  if ((Msg = WM_DESTROY) and (Wnd = ghAlignWindow)) then begin
{$IFDEF DEBUG}
    StrFmt(szODS, 'window %4xh destroyed'#10, [ghAlignWindow]);
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_DEBUGOUT;
      pEvt^.atomMsg:= GlobalAddAtom(szODS);
      AddEvent(pEvt);
    end;
{$ENDIF}
    ghAlignWindow:= 0;
  end;
    { adjust height }
  if ((Msg = WM_GETMINMAXINFO) and (lParam <> 0)) then begin
    wDelta:= GetSystemMetrics(SM_CYCAPTION) - GetSystemMetrics(SM_CYSMCAPTION) - 1;
    with PMinMaxInfo(lParam)^ do begin
      Dec(ptMaxSize.y, wDelta);
      Dec(ptMinTrackSize.y, wDelta);
      Dec(ptMaxTrackSize.y, wDelta);
    end;
  end;
end;

{***}

function LoadLibraryProc(LibFileName: PChar): THandle; export;
var pEvt: PNFYEvent;
begin
  asm
    MOV    AX, SEG @DATA
    MOV    DS, AX
  end;

  ProcUnhook(hookLoadLibrary);
  Result:= LoadLibrary(LibFileName);
  ProcHook(hookLoadLibrary);

  if (GetCurrentTask = ghTaskDelphi) then begin
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_LOADLIBRARY;
      pEvt^.hMod:= Result;
      pEvt^.atomMsg:= GlobalAddAtom(LibFileName);
      AddEvent(pEvt);
    end;
  end;
end;

procedure FreeLibraryProc(LibModule: THandle); export;
var pEvt: PNFYEvent;
begin
  asm
    MOV    AX, SEG @DATA
    MOV    DS, AX
  end;

  ProcUnhook(hookFreeLibrary);
  FreeLibrary(LibModule);
  ProcHook(hookFreeLibrary);

  if (GetCurrentTask = ghTaskDelphi) then begin
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_FREELIBRARY;
      pEvt^.hMod:= LibModule;
      AddEvent(pEvt);
    end;
  end;
end;

{***}

procedure APIInterceptionStart;
begin
    { LoadLibrary() }
  lpfnLoadLibrary:= MakeProcInstance(@LoadLibraryProc, HInstance);
  hookLoadLibrary:= SetProcAddress(@LoadLibrary, lpfnLoadLibrary, FALSE);
    { FreeLibrary() }
  lpfnFreeLibrary:= MakeProcInstance(@FreeLibraryProc, HInstance);
  hookFreeLibrary:= SetProcAddress(@FreeLibrary, lpfnFreeLibrary, FALSE);
    { CreateWindowEx() }
  lpfnCreateWndEx:= MakeProcInstance(@CreateWndExProc, HInstance);
  hookCreateWndEx:= SetProcAddress(@CreateWindowEx, lpfnCreateWndEx, FALSE);

  lpfnNewWndProc:= MakeProcInstance(@AlignWindowProc, HInstance);
end;

procedure APIInterceptionStop;
{$IFDEF DEBUG}
var pEvt: PNFYEvent;
{$ENDIF}
begin
    { if prog gets killed while DELPHI is running restore old
      window procedure of TAlignPalette handle }
  if ((ghAlignWindow <> 0) and IsWindow(ghAlignWindow)) then begin
    SetWindowLong(ghAlignWindow, GWL_WNDPROC, longint(lpfnOldWndProc));
{$IFDEF DEBUG}
    StrFmt(szODS, 'window %4xh: restored window proc from %p to %p'#10,
                  [ghAlignWindow, lpfnNewWndProc, lpfnOldWndProc]);
    New(pEvt);
    if Assigned(pEvt) then begin
      pEvt^.wID:= NFY_MH_DEBUGOUT;
      pEvt^.atomMsg:= GlobalAddAtom(szODS);
      AddEvent(pEvt);
    end;
{$ENDIF}
    ghAlignWindow:= 0;
  end;
  if Assigned(lpfnNewWndProc) then FreeProcInstance(lpfnNewWndProc);
    { CreateWindowEx() }
  if (hookCreateWndEx <> 0) then SetProcRelease(hookCreateWndEx);
  if Assigned(lpfnCreateWndEx) then FreeProcInstance(lpfnCreateWndEx);
    { FreeLibrary() }
  if (hookFreeLibrary <> 0) then SetProcRelease(hookFreeLibrary);
  if Assigned(lpfnFreeLibrary) then FreeProcInstance(lpfnFreeLibrary);
    { LoadLibrary() }
  if (hookLoadLibrary <> 0) then SetProcRelease(hookLoadLibrary);
  if Assigned(lpfnLoadLibrary) then FreeProcInstance(lpfnLoadLibrary);
end;

{$ENDIF} {WIN32}
end.

